"""Axo Order Book Module."""

import json
import logging
import os
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any
from typing import ClassVar
from typing import Union

import requests
from dotenv import load_dotenv
from pycardano import Address
from pycardano import AlonzoMetadata
from pycardano import AuxiliaryData
from pycardano import Metadata
from pycardano import MultiAsset
from pycardano import PlutusData
from pycardano import PlutusV2Script
from pycardano import Redeemer
from pycardano import RedeemerTag
from pycardano import TransactionBuilder
from pycardano import TransactionId
from pycardano import TransactionInput
from pycardano import TransactionOutput
from pycardano import UTxO
from pycardano import Value
from pycardano.utils import min_lovelace
from pydantic import BaseModel
from pydantic import field_validator

from charli3_dendrite.backend import get_backend
from charli3_dendrite.dataclasses.datums import AssetClass
from charli3_dendrite.dataclasses.datums import CancelRedeemer
from charli3_dendrite.dataclasses.datums import OrderDatum
from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import OrderType
from charli3_dendrite.dataclasses.models import PoolSelector
from charli3_dendrite.dexs.core.errors import InvalidPoolError
from charli3_dendrite.dexs.ob.ob_base import AbstractOrderBookState
from charli3_dendrite.dexs.ob.ob_base import BuyOrderBook
from charli3_dendrite.dexs.ob.ob_base import OrderBookOrder
from charli3_dendrite.dexs.ob.ob_base import SellOrderBook
from charli3_dendrite.utility import asset_to_value

formatter = logging.Formatter(
    fmt="%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
)
logger = logging.getLogger("cardem.api.dataclasses.axo")

load_dotenv()

AXO_API_KEY = os.environ.get("AXO_API_KEY")
HTTP_OK = 200
MIN_UTXO_VALUE = 1200000


@dataclass
class TimeMilliseconds(PlutusData):
    """Time in milliseconds for Plutus data."""

    CONSTR_ID = 7
    time_milliseconds: int


@dataclass
class Rationale(PlutusData):
    """Rationale for Plutus data."""

    CONSTR_ID = 0
    numerator: int
    denominator: int


@dataclass
class RationaleWrapper(PlutusData):
    """Wrapper for Rationale Plutus data."""

    CONSTR_ID = 2
    wrapper: Rationale


@dataclass
class AxoOrderDatum(OrderDatum):
    """Axo Order Datum."""

    CONSTR_ID = 0

    node_allocation: dict[int, dict[bytes, dict[bytes, int]]]
    asset_mapping: list[AssetClass]
    instance_token: AssetClass
    parameters: dict[bytes, Union[RationaleWrapper, TimeMilliseconds]]
    variables: dict[Any, Any]

    def address_source(self, block_time: int | None) -> str:
        """Get the address source."""
        if block_time is None:
            raise ValueError("Block time cannot be None")
        address = get_backend().get_axo_target(
            assets=self.instance_token.assets,
            block_time=datetime.fromtimestamp(block_time),
        )
        return Address.decode(address)

    def order_type(self) -> OrderType:
        """Get the order type."""
        return OrderType.swap

    def requested_amount(self) -> Assets:
        """Get the requested amount."""
        tokens = []
        for i, token in self.node_allocation.items():
            if len(token) == 0:
                tokens.append(0)
            else:
                tokens.append(
                    token[self.asset_mapping[i].policy][
                        self.asset_mapping[i].asset_name
                    ],
                )

        price: Rationale | None = None
        for _, value in self.parameters.items():
            if isinstance(value, RationaleWrapper):
                price = value.wrapper

        if price is None:
            raise ValueError("Could not find price")

        unit = (self.asset_mapping[1].policy + self.asset_mapping[1].asset_name).hex()
        if unit == "":
            unit = "lovelace"
        quantity = tokens[1] + tokens[0] * price.denominator // price.denominator
        return Assets(**{unit: quantity})


@dataclass
class AxoCancelRedeemer(PlutusData):
    """Axo Cancel Redeemer."""

    CONSTR_ID = 6


class AxoOBResponse(BaseModel):
    """Axo Order Book Response."""

    amount_unit: str
    amount_unit_ticker: str
    arrow_pair: str
    buy_side_amount: list[float]
    buy_side_depth: int | None = None
    buy_side_price: list[float]
    left: str
    left_ticker: str
    pair: str
    right: str
    right_ticker: str
    sell_side_amount: list[float]
    sell_side_depth: int | None = None
    sell_side_price: list[float]


class AxoCreateParams(BaseModel):
    """Axo Create Parameters."""

    left: str
    right: str
    amount: float | str
    start_date: datetime | None = None
    end_date: datetime | None = None
    price: float | str | None = None
    slippage: float | str = 3.0


class AxoCreateResponse(BaseModel):
    """Axo Create Response."""

    policy_script: str
    strat_id: str
    token_name: str
    datum: str
    algo_addr: str
    nft_metadata: dict

    @field_validator("nft_metadata", mode="before")
    @classmethod
    def validate_metadata(cls, v: str) -> dict:
        """Validate metadata."""
        return json.loads(v)

    @field_validator("policy_script")
    @classmethod
    def strip_policy(cls, v: str) -> str:
        """Trim the extra bytes cbor tag off it."""
        return v[6:]


class AxoCloseResponse(BaseModel):
    """Axo Close Response."""

    validator_address: str
    command_datum: str


class AxoCMCResponse(BaseModel):
    """Axo CMC Response."""

    base_currency: str
    base_subject: str
    base_volume: float
    highest_bid: float
    highest_price_24h: float | None
    last_price: float | None
    lowest_ask: float
    lowest_price_24h: float | None
    price_change_percent_24h: float | None
    quote_currency: str
    quote_subject: str
    quote_volume: float
    trading_pairs: str


class AxoAlgoName(Enum):
    """Axo Algorithm Names."""

    limit = "Limit"
    market = "Smart Market"
    dca = "DCA"


class AxoAPIClient:
    """Axo API Client."""

    def __init__(self) -> None:
        """Initialize Axo API Client."""
        self.headers = {"Content-Type": "application/json"}
        if AXO_API_KEY is not None:
            self.headers["x-api-key"] = AXO_API_KEY

    network = "mainnet"

    urls: ClassVar[dict[str, str]] = {
        "mainnet": "https://api.axo.trade/",
        "preprod": "https://api.axo-preview.trade/",
    }

    def cmc_summary(self) -> list[AxoCMCResponse]:
        """Get CMC summary."""
        url = self.urls[self.network] + "cmc/summary"

        result = requests.get(url, headers=self.headers, timeout=10)

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return [AxoCMCResponse.model_validate(token) for token in result.json()]

    def aob(self, token_a: str, token_b: str) -> AxoOBResponse:
        """Get AOB."""
        url = self.urls[self.network] + "aob"

        token_a_param = "" if token_a == "lovelace" else token_a  # noqa: S105
        token_b_param = "" if token_b == "lovelace" else token_b  # noqa: S105

        result = requests.get(
            url,
            headers=self.headers,
            params={"left": token_a_param, "right": token_b_param},
            timeout=10,
        )

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return AxoOBResponse.model_validate(result.json())

    def ob(self, token_a: str, token_b: str) -> AxoOBResponse:
        """Get OB."""
        url = self.urls[self.network] + "ob"

        token_a_param = "" if token_a == "lovelace" else token_a  # noqa: S105
        token_b_param = "" if token_b == "lovelace" else token_b  # noqa: S105

        result = requests.get(
            url,
            headers=self.headers,
            params={"left": token_a_param, "right": token_b_param},
            timeout=10,
        )

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return AxoOBResponse.model_validate(result.json())

    def spot(self, token_a: str, token_b: str) -> float | None:
        """Get spot price."""
        url = self.urls[self.network] + "spot"

        token_a_param = "" if token_a == "lovelace" else token_a  # noqa: S105
        token_b_param = "" if token_b == "lovelace" else token_b  # noqa: S105

        result = requests.get(
            url,
            headers=self.headers,
            params={"left": token_a_param, "right": token_b_param},
            timeout=10,
        )

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return result.json()

    def create(
        self,
        wallet_addr: str,
        tx_hash: str,
        tx_idx: int,
        params: AxoCreateParams,
        algo_name: AxoAlgoName = AxoAlgoName.market,
    ) -> AxoCreateResponse:
        """Create Axo order."""
        url = self.urls[self.network] + "create"

        result = requests.post(
            url,
            headers=self.headers,
            json={
                "wallet_addr": wallet_addr,
                "outref_utxo_id": tx_hash,
                "outref_utxo_ix": tx_idx,
                "algo_name": algo_name.value,
                "params": params.model_dump(exclude_none=True),
            },
            timeout=10,
        )

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return AxoCreateResponse.model_validate(result.json())

    def notify(self, tx_hash: str, strat_id: str) -> requests.Response:
        """Notify Axo of transaction."""
        url = self.urls[self.network] + "notify"

        return requests.put(
            url,
            headers=self.headers,
            json={"tx_id": tx_hash, "strat_id": strat_id},
            timeout=10,
        )

    def close(
        self,
        wallet_address: str,
        return_address: str,
        strategy_id: str,
    ) -> AxoCloseResponse:
        """Close Axo order."""
        url = self.urls[self.network] + "close"

        result = requests.post(
            url,
            headers=self.headers,
            json={
                "wallet_address": wallet_address,
                "return_address": return_address,
                "strategy_id": strategy_id,
            },
            timeout=10,
        )

        if result.status_code != HTTP_OK:
            raise ValueError(f"API request failed: {result.status_code}: {result.text}")

        return AxoCloseResponse.model_validate(result.json())

    def get_ob_info(
        self,
        assets: Assets,
    ) -> tuple[AxoOBResponse, AxoOBResponse, float | None]:
        """Get order book info."""
        return (
            self.aob(
                token_a=assets.unit(0),
                token_b=assets.unit(1),
            ),
            self.ob(
                token_a=assets.unit(0),
                token_b=assets.unit(1),
            ),
            self.spot(
                token_a=assets.unit(0),
                token_b=assets.unit(1),
            ),
        )


class AxoOBMarketState(AbstractOrderBookState):
    """Axo Order Book Market State."""

    fee: int = 10
    spot: float = 1
    plutus_v2: bool = True
    inactive: bool = False
    _stake_address: ClassVar[Address] = Address.decode(
        "addr1z92l7rnra7sxjn5qv5fzc4fwsrrm29mgkleqj9a0y46j5lrryf9mtf9layje8u7u7wmap6alr28l90ry5t9nlyldjjsse4mxc9",
    )
    _client: ClassVar[AxoAPIClient] = AxoAPIClient()
    _reference_utxo: ClassVar[UTxO | None] = None
    _deposit: Assets = Assets(lovelace=8000000)

    @classmethod
    def dex(cls) -> str:
        """Get DEX name."""
        return "Axo"

    @classmethod
    def order_selector(cls) -> list[str]:
        """Order selection information."""
        return [
            "addr1z92l7rnra7sxjn5qv5fzc4fwsrrm29mgkleqj9a0y46j5lrryf9mtf9layje8u7u7wmap6alr28l90ry5t9nlyldjjsse4mxc9",
        ]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        """Pool selection information."""
        return []

    @property
    def swap_forward(self) -> bool:
        """Check if swap is forward."""
        return False

    @classmethod
    def default_script_class(cls) -> type[PlutusV2Script]:
        """Get default script class."""
        return PlutusV2Script

    @classmethod
    def reference_utxo(cls) -> UTxO | None:
        """Get reference UTxO."""
        if cls._reference_utxo is None:
            script_reference = get_backend().get_script_from_address(cls._stake_address)

            script = cls.default_script_class()(bytes.fromhex(script_reference.script))

            cls._reference_utxo = UTxO(
                input=TransactionInput(
                    transaction_id=TransactionId(
                        bytes.fromhex(
                            script_reference.tx_hash,
                        ),
                    ),
                    index=script_reference.tx_index,
                ),
                output=TransactionOutput(
                    address=Address.decode(script_reference.address),
                    amount=asset_to_value(script_reference.assets),
                    script=script,
                ),
            )
        return cls._reference_utxo

    @classmethod
    def _process_ob(
        cls,
        ob: AxoOBResponse,
    ) -> tuple[list[OrderBookOrder], list[OrderBookOrder]]:
        """Process order book response."""
        prices = get_token_prices(assets=[ob.left, ob.right])  # type: ignore # noqa: F821
        left = ob.left if ob.left != "" else "lovelace"
        if prices[0].policy_id + prices[0].policy_name == left:
            token_a_decimals = prices[0].decimals
            token_b_decimals = prices[1].decimals
        else:
            token_a_decimals = prices[1].decimals
            token_b_decimals = prices[0].decimals

        sell_book = []
        for index in range(len(ob.sell_side_price)):
            sell_book.append(
                OrderBookOrder(
                    price=ob.sell_side_price[index]
                    * 10 ** (token_a_decimals - token_b_decimals),
                    quantity=int(ob.sell_side_amount[index] * 10**token_b_decimals),
                ),
            )
        buy_book = []
        for index in range(len(ob.buy_side_price)):
            buy_book.append(
                OrderBookOrder(
                    price=ob.buy_side_price[index] ** -1
                    * 10 ** (token_b_decimals - token_a_decimals),
                    quantity=int(
                        ob.buy_side_amount[index]
                        * 10**token_a_decimals
                        * ob.buy_side_price[index],
                    ),
                ),
            )

        return BuyOrderBook(buy_book), SellOrderBook(sell_book)

    @classmethod
    def get_book(cls, assets: Assets) -> "AxoOBMarketState":
        """Get order book."""
        aob, ob, spot = cls._client.get_ob_info(assets)

        if spot is None:
            raise InvalidPoolError("Spot price is None")

        try:
            buy_book, sell_book = cls._process_ob(ob=aob)
            buy_book_full, sell_book_full = cls._process_ob(ob=ob)
        except IndexError as err:
            logger.error(f"Error getting Axo order book for assets: {assets}")
            raise InvalidPoolError("Error processing order book") from err

        if "lovelace" in assets:
            spot = 1.0

        return cls(
            assets=assets,
            spot=spot,
            block_time=int(datetime.now().timestamp()),
            block_index=0,
            buy_book_full=buy_book_full,
            sell_book_full=sell_book_full,
        )

    @classmethod
    def order_datum_class(cls) -> type[PlutusData]:
        """Get order datum class."""
        return AxoOrderDatum

    @property
    def stake_address(self) -> Address:
        """Get stake address."""
        return self._stake_address

    def swap_utxo(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        tx_builder: TransactionBuilder,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> tuple[TransactionOutput, PlutusData, UTxO]:
        """Swap UTxO."""
        # Basic checks
        if len(in_assets) != 1 or len(out_assets) != 1:
            raise ValueError(
                "Only one asset can be supplied as input, "
                + "and one asset supplied as output.",
            )

        # Get the mint input UTxO
        utxo_input = None
        for utxo in tx_builder.inputs:
            if utxo.output.amount.coin > MIN_UTXO_VALUE:  # noqa: SIM102
                if utxo_input is None or len(utxo_input.output.to_cbor_hex()) < len(
                    utxo.output.to_cbor_hex(),
                ):
                    utxo_input = utxo

        if utxo_input is None:
            raise ValueError("No suitable input UTxO found")

        # Get the order build info
        prices = get_token_prices(assets=[in_assets.unit(), out_assets.unit()])  # type: ignore # noqa: F821
        if prices[0].policy_id + prices[0].policy_name == in_assets.unit():
            in_decimals = prices[0].decimals
        else:
            in_decimals = prices[1].decimals

        params = AxoCreateParams(
            left=in_assets.unit() if in_assets.unit() != "lovelace" else "",
            right=out_assets.unit() if out_assets.unit() != "lovelace" else "",
            amount=in_assets.quantity() / 10**in_decimals,
            slippage=self.slippage(in_assets=in_assets, out_assets=out_assets) + 2.0,
        )
        create: AxoCreateResponse = self._client.create(
            wallet_addr=address_source.encode(),
            tx_hash=utxo_input.input.transaction_id.payload.hex(),
            tx_idx=utxo_input.input.index,
            params=params,
        )

        # Create the mint metadata
        metadata = AuxiliaryData(
            AlonzoMetadata(metadata=Metadata({721: create.nft_metadata})),
        )
        tx_builder.auxiliary_data = metadata

        # Create the axo receipt
        axo_receipt = MultiAsset.from_primitive(
            {
                bytes.fromhex(create.strat_id): {
                    bytes.fromhex(create.token_name): 2,
                },
            },
        )
        tx_builder.mint = axo_receipt
        redeemer = Redeemer(CancelRedeemer())
        redeemer.tag = RedeemerTag.MINT
        tx_builder.add_minting_script(
            PlutusV2Script(bytes.fromhex(create.policy_script)),
            redeemer,
        )

        # Add in the recommended lovelace to the input and the axo receipt
        in_assets.root["lovelace"] = (
            in_assets["lovelace"]
            + self.batcher_fee(
                in_assets=in_assets,
                out_assets=out_assets,
                extra_assets=extra_assets,
            ).quantity()
            + self.deposit(in_assets=in_assets, out_assets=out_assets).quantity()
        )
        in_assets.root[create.strat_id + create.token_name] = 1

        # Create the swap utxo
        order_datum = AxoOrderDatum.from_cbor(create.datum)
        output = TransactionOutput(
            address=create.algo_addr,
            amount=asset_to_value(in_assets),
            datum=order_datum,
        )
        tx_builder.add_output(output)

        # Create the receipt UTxO
        utxo = TransactionOutput(
            address=address_source,
            amount=Value(
                coin=1000000,
                multi_asset=MultiAsset.from_primitive(
                    {
                        bytes.fromhex(create.strat_id): {
                            bytes.fromhex(create.token_name): 1,
                        },
                    },
                ),
            ),
        )
        utxo.amount.coin = min_lovelace(context=tx_builder.context, output=utxo)
        tx_builder.add_output(utxo)

        return output, order_datum, utxo_input

    @property
    def volume_fee(self) -> int:
        """Get volume fee."""
        return 10

    @classmethod
    def cancel_redeemer(cls) -> PlutusData:
        """Get cancel redeemer."""
        return Redeemer(AxoCancelRedeemer())

    def batcher_fee(
        self,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
    ) -> Assets:
        """Batcher fee.

        Args:
            in_assets: The input assets for the swap
            out_assets: The output assets for the swap
            extra_assets: Extra assets included in the transaction
        """
        if in_assets.unit() == "lovelace":
            fees = max(self.volume_fee * in_assets.quantity() // 10000, 1200000)
        elif out_assets.unit() == "lovelace":
            fees = max(self.volume_fee * out_assets.quantity() // 10000, 1200000)
        else:
            fees = max(
                self.volume_fee * in_assets.quantity() * self.spot // 10000,
                1200000,
            )

        # The below code estimates the Cardano cost of executing the tx
        fees += 250000  # ~cost of output tx

        if in_assets.unit() == self.unit_a:
            book = self.sell_book_full
        else:
            book = self.buy_book_full

        # Each fill order incurs ~0.6 ada cost
        index = 0
        in_quantity = in_assets.quantity()
        while in_quantity > 0 and index < len(book):
            available = book[index].quantity * book[index].price
            fees += 600000
            if available > in_quantity:
                in_quantity = 0
            else:
                in_quantity -= book[index].price * book[index].quantity
            index += 1

        return Assets(lovelace=fees)

    def slippage(
        self,
        in_assets: Assets,
        out_assets: Assets,
    ) -> float:
        """Calculate slippage.

        Args:
            in_assets: The input assets for the swap
            out_assets: The output assets for the swap
        """
        if in_assets.unit() == "lovelace":
            fees = max(self.volume_fee * in_assets.quantity() // 10000, 1200000)
        elif out_assets.unit() == "lovelace":
            fees = max(self.volume_fee * out_assets.quantity() // 10000, 1200000)
        else:
            fees = max(
                self.volume_fee * in_assets.quantity() * self.spot // 10000,
                1200000,
            )

        # The below code estimates the Cardano cost of executing the tx
        fees += 250000  # ~cost of output tx

        if in_assets.unit() == self.unit_a:
            book = self.sell_book_full
        else:
            book = self.buy_book_full

        # Each fill order incurs ~0.5 ada cost
        index = 0
        best_price = book[index].price if book else 0
        in_quantity = in_assets.quantity()
        last_price = best_price
        while in_quantity > 0 and index < len(book):
            available = book[index].quantity * book[index].price
            fees += 500000
            if available > in_quantity:
                in_quantity = 0
            else:
                in_quantity -= book[index].price * book[index].quantity
            last_price = book[index].price
            index += 1

        return 100 * abs(1 - (best_price / last_price)) if last_price != 0 else 0

    @property
    def pool_id(self) -> str:
        """Get the pool ID."""
        return f"{self.dex}.{self.unit_a}.{self.unit_b}"



"""Abstract base class for Cardano blockchain backend implementations."""

from abc import ABC
from abc import abstractmethod
from datetime import datetime
from typing import Optional

from pycardano import Address  # type: ignore

from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import BlockList
from charli3_dendrite.dataclasses.models import PoolStateList
from charli3_dendrite.dataclasses.models import ScriptReference
from charli3_dendrite.dataclasses.models import SwapTransactionList


class AbstractBackend(ABC):
    """Abstract base class for Cardano blockchain backend implementations.

    This class defines the interface for interacting with various Cardano blockchain
    data sources such as db-sync, Ogmios, or Kupo.
    """

    @abstractmethod
    def get_pool_utxos(
        self,
        addresses: list[str],
        assets: list[str] | None = None,
        limit: int = 1000,
        page: int = 0,
        historical: bool = True,
    ) -> PoolStateList:
        """Get UTXOs for specific assets or addresses.

        Args:
            assets (Optional[List[str]]): List of asset IDs to filter by.
            addresses (Optional[List[str]]): List of addresses to filter by.
            limit (int): Maximum number of results to return.
            page (int): Page number for pagination.
            historical (bool): Whether to include historical data.

        Returns:
            PoolStateList: List of pool state objects.
        """
        pass

    @abstractmethod
    def get_pool_in_tx(
        self,
        tx_hash: str,
        addresses: list[str],
        assets: list[str] | None = None,
    ) -> PoolStateList:
        """Get pool state for a specific transaction.

        Args:
            tx_hash (str): The transaction hash to query.
            assets (Optional[List[str]]): List of asset IDs to filter by.
            addresses (Optional[List[str]]): List of addresses to filter by.

        Returns:
            PoolStateList: List of pool state objects for the transaction.
        """
        pass

    @abstractmethod
    def last_block(self, last_n_blocks: int = 2) -> BlockList:
        """Get information about the last n blocks.

        Args:
            last_n_blocks (int): Number of recent blocks to retrieve.

        Returns:
            BlockList: List of recent block information.
        """
        pass

    @abstractmethod
    def get_pool_utxos_in_block(self, block_no: int) -> PoolStateList:
        """Get pool UTXOs for a specific block.

        Args:
            block_no (int): The block number to query.

        Returns:
            PoolStateList: List of pool state objects for the block.
        """
        pass

    @abstractmethod
    def get_script_from_address(self, address: Address) -> ScriptReference:
        """Get script reference for a given address.

        Args:
            address (Address): The address to query.

        Returns:
            ScriptReference: Script reference for the address.
        """
        pass

    @abstractmethod
    def get_historical_order_utxos(
        self,
        stake_addresses: list[str],
        after_time: datetime | int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get historical order UTXOs for given stake addresses.

        Args:
            stake_addresses (List[str]): List of stake addresses to query.
            after_time (Optional[Union[datetime, int]]): Filter results after this time.
            limit (int): Maximum number of results to return.
            page (int): Page number for pagination.

        Returns:
            SwapTransactionList: List of swap transaction objects.
        """
        pass

    @abstractmethod
    def get_order_utxos_by_block_or_tx(
        self,
        stake_addresses: list[str],
        out_tx_hash: list[str] | None = None,
        in_tx_hash: list[str] | None = None,
        block_no: int | None = None,
        after_block: int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get order UTXOs by block or transaction.

        Args:
            stake_addresses (List[str]): List of stake addresses to query.
            out_tx_hash (Optional[List[str]]): List of transaction hashes to filter by.
            in_tx_hash: list of input transaction hashes to filter by.
            block_no (Optional[int]): Specific block number to query.
            after_block (Optional[int]): Filter results after this block number.
            limit (int): Maximum number of results to return.
            page (int): Page number for pagination.

        Returns:
            SwapTransactionList: List of swap transaction objects.
        """
        pass

    @abstractmethod
    def get_cancel_utxos(
        self,
        stake_addresses: list[str],
        block_no: int | None = None,
        after_time: datetime | int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get cancelled order UTXOs.

        Args:
            stake_addresses (List[str]): List of stake addresses to query.
            block_no (Optional[int]): Specific block number to query.
            after_time (Optional[Union[datetime, int]]): Filter results after this time.
            limit (int): Maximum number of results to return.
            page (int): Page number for pagination.

        Returns:
            SwapTransactionList: List of swap transaction objects for cancelled orders.
        """
        pass

    @abstractmethod
    def get_datum_from_address(
        self,
        address: str,
        asset: str | None = None,
    ) -> Optional[ScriptReference]:
        """Get datum from a given address.

        Args:
            address: The address to query.
            asset: Assets required to be in the UTxO.

        Returns:
            The datum associated with the address, if any.
        """
        pass

    @abstractmethod
    def get_axo_target(
        self,
        assets: Assets,
        block_time: datetime | None = None,
    ) -> str | None:
        """Get the target address for the given assets.

        Args:
            assets: The assets to query.
            block_time: The block time to query.

        Returns:
            The target address for the assets, if any.
        """
        pass



"""Module providing base classes for OB pools."""

from abc import abstractmethod
from decimal import Decimal
from math import ceil

from pycardano import DeserializeException
from pycardano import PlutusData
from pycardano import UTxO
from pydantic import model_validator

from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import BaseList
from charli3_dendrite.dataclasses.models import DendriteBaseModel
from charli3_dendrite.dexs.core.base import AbstractPairState
from charli3_dendrite.dexs.core.errors import InvalidPoolError
from charli3_dendrite.dexs.core.errors import NoAssetsError
from charli3_dendrite.dexs.core.errors import NotAPoolError

ASSET_COUNT_ONE = 1
ASSET_COUNT_TWO = 2
ASSET_COUNT_THREE = 3


class AbstractOrderState(AbstractPairState):
    """This class is largely used for OB dexes that allow direct script inputs."""

    tx_hash: str
    tx_index: int
    datum_cbor: str
    datum_hash: str
    inactive: bool = False

    _batcher_fee: Assets
    _datum_parsed: PlutusData | None = None

    @property
    def in_unit(self) -> str:
        return self.assets.unit()

    @property
    def out_unit(self) -> str:
        return self.assets.unit(1)

    @property
    @abstractmethod
    def price(self) -> tuple[int, int]:
        raise NotImplementedError

    @property
    @abstractmethod
    def available(self) -> Assets:
        """Max amount of output asset that can be used to fill the order."""
        raise NotImplementedError

    def get_amount_out(self, asset: Assets, precise=True) -> tuple[Assets, float]:
        assert asset.unit() == self.in_unit and len(asset) == 1

        num, denom = self.price
        out_assets = Assets(**{self.out_unit: 0})
        in_quantity = asset.quantity() - ceil(
            asset.quantity() * self.volume_fee / 10000,
        )
        out_assets.root[self.out_unit] = min(
            ceil(in_quantity * denom / num),
            self.available.quantity(),
        )

        if precise:
            out_assets.root[self.out_unit] = int(out_assets.quantity())

        return out_assets, 0

    def get_amount_in(self, asset: Assets, precise=True) -> tuple[Assets, float]:
        assert asset.unit() == self.out_unit and len(asset) == 1

        denom, num = self.price
        in_assets = Assets(**{self.in_unit: 0})
        out_quantity = asset.quantity()
        in_assets.root[self.in_unit] = (
            min(out_quantity, self.available.quantity()) * denom
        ) / num
        fees = in_assets[self.in_unit] * self.volume_fee / 10000
        in_assets.root[self.in_unit] += fees

        if precise:
            in_assets.root[self.in_unit] = int(in_assets.quantity())

        return in_assets, 0

    @classmethod
    def skip_init(cls, values: dict[str, ...]) -> bool:
        """An initial check to determine if parsing should be carried out.

        Args:
            values: The pool initialization parameters.

        Returns:
            bool: If this returns True, initialization checks will get skipped.
        """
        return False

    @classmethod
    def extract_dex_nft(cls, values: dict[str, ...]) -> Assets | None:
        """Extract the dex nft from the UTXO.

        Some DEXs put a DEX nft into the pool UTXO.

        This function checks to see if the DEX nft is in the UTXO if the DEX policy is
        defined.

        If the dex nft is in the values, this value is skipped because it is assumed
        that this utxo has already been parsed.

        Args:
            values: The pool UTXO inputs.

        Returns:
            Assets: None or the dex nft.
        """
        assets = values["assets"]

        # If no dex policy id defined, return nothing
        if cls.dex_policy() is None:
            dex_nft = None

        # If the dex nft is in the values, it's been parsed already
        elif "dex_nft" in values:
            if not any(
                any(p.startswith(d) for d in cls.dex_policy())
                for p in values["dex_nft"]
            ):
                raise NotAPoolError("Invalid DEX NFT")
            dex_nft = values["dex_nft"]

        # Check for the dex nft
        else:
            nfts = [
                asset
                for asset in assets
                if any(asset.startswith(policy) for policy in cls.dex_policy())
            ]
            if len(nfts) < 1:
                raise NotAPoolError(
                    f"{cls.__name__}: Pool must have one DEX NFT token.",
                )
            dex_nft = Assets(**{nfts[0]: assets.root.pop(nfts[0])})
            values["dex_nft"] = dex_nft

        return dex_nft

    @property
    def order_datum(self) -> PlutusData:
        if self._datum_parsed is None:
            self._datum_parsed = self.order_datum_class().from_cbor(self.datum_cbor)
        return self._datum_parsed

    @classmethod
    def post_init(cls, values: dict[str, ...]):
        """Post initialization checks.

        Args:
            values: The pool initialization parameters
        """
        assets = values["assets"]
        non_ada_assets = [a for a in assets if a != "lovelace"]

        if len(assets) == 2:
            # ADA pair
            assert (
                len(non_ada_assets) == 1
            ), f"Pool must only have 1 non-ADA asset: {values}"

        elif len(assets) == 3:
            # Non-ADA pair
            assert len(non_ada_assets) == 2, "Pool must only have 2 non-ADA assets."

            # Send the ADA token to the end
            values["assets"].root["lovelace"] = values["assets"].root.pop("lovelace")

        else:
            if len(assets) == 1 and "lovelace" in assets:
                raise NoAssetsError(
                    f"Invalid pool, only contains lovelace: assets={assets}",
                )
            else:
                raise InvalidPoolError(
                    f"Pool must have 2 or 3 assets except factor, NFT, and LP tokens: assets={assets}",
                )
        return values

    @model_validator(mode="before")
    def translate_address(cls, values):
        """The main validation function called when initialized.

        Args:
            values: The pool initialization values.

        Returns:
            The parsed/modified pool initialization values.
        """
        if "assets" in values:
            if values["assets"] is None:
                raise NoAssetsError("No assets in the pool.")
            elif not isinstance(values["assets"], Assets):
                values["assets"] = Assets(**values["assets"])

        if cls.skip_init(values):
            return values

        # Parse the order datum
        try:
            datum = cls.order_datum_class().from_cbor(values["datum_cbor"])
        except (DeserializeException, TypeError) as e:
            raise NotAPoolError(
                "Order datum could not be deserialized: \n "
                + f"    error={e}\n"
                + f"    tx_hash={values['tx_hash']}\n"
                + f"    datum={values['datum_cbor']}\n",
            )

        # To help prevent edge cases, remove pool tokens while running other checks
        pair = datum.pool_pair()
        if datum.pool_pair() is not None:
            for token in datum.pool_pair():
                try:
                    if token in values["assets"]:
                        pair.root.update({token: values["assets"].root.pop(token)})
                except KeyError:
                    raise InvalidPoolError(
                        "Order does not contain expected asset.\n"
                        + f"    Expected: {token}\n"
                        + f"    Actual: {values['assets']}",
                    )

        dex_nft = cls.extract_dex_nft(values)

        # Add the pool tokens back in
        values["assets"].root.update(pair.root)

        cls.post_init(values)

        return values


class OrderBookOrder(DendriteBaseModel):
    """Represents an order in the order book."""

    price: float
    quantity: int
    state: AbstractOrderState | None = None


class BuyOrderBook(BaseList):
    root: list[OrderBookOrder]

    @model_validator(mode="after")
    def sort_descend(self):
        self.root.sort(key=lambda x: x.price)
        return self


class SellOrderBook(BaseList):
    root: list[OrderBookOrder]

    @model_validator(mode="after")
    def sort_descend(self):
        self.root.sort(key=lambda x: x.price)
        return self


class AbstractOrderBookState(AbstractPairState):
    """This class is largely used for OB dexes that have a batcher."""

    sell_book: SellOrderBook | None = None
    buy_book: BuyOrderBook | None = None
    sell_book_full: SellOrderBook
    buy_book_full: BuyOrderBook

    def get_amount_out(
        self,
        asset: Assets,
        precise: bool = True,
        apply_fee: bool = False,
    ) -> tuple[Assets, float]:
        """Get the amount of token output for the given input.

        Args:
            asset: The input assets
            precise: If precise, uses integers. Defaults to True.

        Returns:
            tuple[Assets, float]: The output assets and slippage.
        """
        assert len(asset) == 1, "Asset should only have one token."
        assert asset.unit() in [
            self.unit_a,
            self.unit_b,
        ], f"Asset {asset.unit} is invalid for pool {self.unit_a}-{self.unit_b}"

        if asset.unit() == self.unit_a:
            book = self.sell_book_full
            unit_out = self.unit_b
        else:
            book = self.buy_book_full
            unit_out = self.unit_a

        in_quantity = asset.quantity()
        if apply_fee:
            in_quantity = in_quantity * (10000 - self.fee) // 10000

        index = 0
        out_assets = Assets({unit_out: 0})
        while in_quantity > 0 and index < len(book):
            available = book[index].quantity * book[index].price
            if available > in_quantity:
                out_assets.root[unit_out] += in_quantity / book[index].price
                in_quantity = 0
            else:
                out_assets.root[unit_out] += book[index].quantity
                in_quantity -= book[index].price * book[index].quantity
            index += 1

        out_assets.root[unit_out] = int(out_assets[unit_out])

        return out_assets, 0

    def get_amount_in(
        self,
        asset: Assets,
        precise: bool = True,
        apply_fee: bool = False,
    ) -> tuple[Assets, float]:
        """Get the amount of token input for the given output.

        Args:
            asset: The input assets
            precise: If precise, uses integers. Defaults to True.

        Returns:
            tuple[Assets, float]: The output assets and slippage.
        """
        assert len(asset) == 1, "Asset should only have one token."
        assert asset.unit() in [
            self.unit_a,
            self.unit_b,
        ], f"Asset {asset.unit} is invalid for pool {self.unit_a}-{self.unit_b}"

        if asset.unit() == self.unit_b:
            book = self.sell_book_full
            unit_in = self.unit_a
        else:
            book = self.buy_book_full
            unit_in = self.unit_b

        index = 0
        out_quantity = asset.quantity()
        in_assets = Assets({unit_in: 0})
        while out_quantity > 0 and index < len(book):
            available = book[index].quantity
            if available > out_quantity:
                in_assets.root[unit_in] += out_quantity * book[index].price
                out_quantity = 0
            else:
                in_assets.root[unit_in] += book[index].quantity / book[index].price
                out_quantity -= book[index].quantity
            index += 1

        if apply_fee:
            fees = in_assets[unit_in] * self.fee / 10000
            in_assets.root[unit_in] += fees

        in_assets.root[unit_in] = int(in_assets[unit_in])

        return in_assets, 0

    @classmethod
    def reference_utxo(self) -> UTxO | None:
        return None

    @property
    def price(self) -> tuple[Decimal, Decimal]:
        """Mid price of assets.

        Returns:
            A `Tuple[Decimal, Decimal] where the first `Decimal` is the price to buy
                1 of token B in units of token A, and the second `Decimal` is the price
                to buy 1 of token A in units of token B.
        """
        prices = (
            Decimal((self.buy_book[0].price + 1 / self.sell_book[0].price) / 2),
            Decimal((self.sell_book[0].price + 1 / self.buy_book[0].price) / 2),
        )

        return prices

    @property
    def tvl(self) -> Decimal:
        """Return the total value locked for the pool.

        Raises:
            NotImplementedError: Only ADA pool TVL is implemented.
        """
        if self.unit_a != "lovelace":
            raise NotImplementedError("tvl for non-ADA pools is not implemented.")

        tvl = sum(b.quantity / b.price for b in self.buy_book) + sum(
            s.quantity * s.price for s in self.sell_book
        )

        return Decimal(int(tvl) / 10**6)

    @classmethod
    @abstractmethod
    def get_book(
        cls,
        assets: Assets | None = None,
        orders: list[AbstractOrderState] | None = None,
    ) -> "AbstractOrderBookState":
        raise NotImplementedError



"""Abstract base class and common functions for handling token pairs."""

from abc import ABC
from abc import abstractmethod
from decimal import Decimal

from pycardano import Address
from pycardano import PlutusData
from pycardano import PlutusV1Script
from pycardano import PlutusV2Script
from pycardano import Redeemer
from pycardano import TransactionOutput
from pycardano import UTxO

from charli3_dendrite.dataclasses.datums import CancelRedeemer
from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import DendriteBaseModel
from charli3_dendrite.dataclasses.models import PoolSelector


class AbstractPairState(DendriteBaseModel, ABC):
    """Abstract base class representing the state of a pair."""

    assets: Assets
    block_time: int
    block_index: int
    fee: int | list[int] | None = None
    plutus_v2: bool
    tx_index: int | None = None
    tx_hash: str | None = None
    datum_cbor: str | None = None
    datum_hash: str | None = None
    dex_nft: Assets | None = None

    _batcher_fee: Assets
    _datum_parsed: PlutusData

    @classmethod
    @abstractmethod
    def dex(cls) -> str:
        """Official dex name."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def order_selector(cls) -> list[str]:
        """Order selection information."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def pool_selector(cls) -> PoolSelector:
        """Pool selection information."""
        raise NotImplementedError

    @abstractmethod
    def get_amount_out(self, asset: Assets) -> tuple[Assets, float]:
        """Calculate the output amount of assets for given input.

        Args:
            asset: An asset with a defined quantity.

        Returns:
            A tuple where the first value is the estimated asset returned and
            the second value is the price impact ratio.
        """
        raise NotImplementedError

    @abstractmethod
    def get_amount_in(self, asset: Assets) -> tuple[Assets, float]:
        """Get the input asset amount given a desired output asset amount.

        Args:
            asset: An asset with a defined quantity.

        Returns:
           A tuple where the first value is the the estimated asset needed and
           the second value is the slippage fee.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def swap_forward(self) -> bool:
        """Returns if swap forwarding is enabled."""
        raise NotImplementedError

    @property
    def inline_datum(self) -> bool:
        """Determine whether the datum should be inline."""
        return self.plutus_v2

    @classmethod
    def reference_utxo(cls) -> UTxO | None:
        """Get Reference UTXO.

        Returns:
            UTxO | None: UTxO object if it exists, otherwise None.
        """
        return None

    @property
    @abstractmethod
    def stake_address(self) -> Address:
        """Return the staking address."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def order_datum_class(cls) -> type[PlutusData]:
        """Returns data class used for handling order datums."""
        raise NotImplementedError

    @classmethod
    def default_script_class(cls) -> type[PlutusV1Script] | type[PlutusV2Script]:
        """Get default script class as Plutus V1 unless overridden.

        Returns:
            type[PlutusV1Script] | type[PlutusV2Script]: The default script class.
        """
        return PlutusV1Script

    @property
    def script_class(self) -> type[PlutusV1Script] | type[PlutusV2Script]:
        """Returns the script class based on the Plutus version being used."""
        if self.plutus_v2:
            return PlutusV2Script
        return PlutusV1Script

    def swap_datum(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> PlutusData:
        """Constructs the datum for a swap transaction."""
        raise NotImplementedError

    @abstractmethod
    def swap_utxo(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> TransactionOutput:
        """Constructs the transaction output for a swap."""
        raise NotImplementedError

    @property
    def volume_fee(self) -> int | float | list[int] | list[float] | None:
        """Swap fee of swap in basis points."""
        return self.fee

    @classmethod
    def cancel_redeemer(cls) -> PlutusData:
        """Returns the redeemer data for canceling transaction."""
        return Redeemer(CancelRedeemer())

    def batcher_fee(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
        extra_assets: Assets | None = None,
    ) -> Assets:
        """Batcher fee.

        Args:
            in_assets: The input assets for the swap
            out_assets: The output assets for the swap
            extra_assets: Extra assets included in the transaction
        """
        return self._batcher

    def deposit(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
    ) -> Assets:
        """Batcher fee."""
        return self._deposit

    @classmethod
    def dex_policy(cls) -> list[str] | None:
        """The dex nft policy.

        This should be the policy or policy+name of the dex nft.

        If None, then the default dex nft check is skipped.

        Returns:
            Optional[str]: policy or policy+name of dex nft
        """
        return None

    @property
    def unit_a(self) -> str:
        """Token name of asset A."""
        return self.assets.unit(0)

    @property
    def unit_b(self) -> str:
        """Token name of asset b."""
        return self.assets.unit(1)

    @property
    def reserve_a(self) -> int:
        """Reserve amount of asset A."""
        return self.assets.quantity(0)

    @property
    def reserve_b(self) -> int:
        """Reserve amount of asset B."""
        return self.assets.quantity(1)

    @property
    @abstractmethod
    def price(self) -> tuple[Decimal, Decimal]:
        """Price of assets.

        Returns:
            A `Tuple[Decimal, Decimal] where the first `Decimal` is the price to buy
                1 of token B in units of token A, and the second `Decimal` is the price
                to buy 1 of token A in units of token B.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def tvl(self) -> Decimal:
        """Return the total value locked for the pool.

        Raises:
            NotImplementedError: Only ADA pool TVL is implemented.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def pool_id(self) -> str:
        """A unique identifier for the pool or ob.

        Raises:
            NotImplementedError: Only ADA pool TVL is implemented.
        """
        raise NotImplementedError



"""Concrete implementation of AbstractBackend for db-sync."""
import logging
import os
from datetime import datetime
from threading import Lock

import psycopg_pool  # type: ignore
from dotenv import load_dotenv  # type: ignore
from psycopg.rows import dict_row  # type: ignore
from psycopg_pool import PoolTimeout
from pycardano import Address  # type: ignore

from charli3_dendrite.backend.backend_base import AbstractBackend
from charli3_dendrite.backend.dbsync.models import OrderSelector
from charli3_dendrite.backend.dbsync.models import PoolSelector
from charli3_dendrite.backend.dbsync.models import UTxOSelector
from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import BlockList
from charli3_dendrite.dataclasses.models import PoolStateList
from charli3_dendrite.dataclasses.models import ScriptReference
from charli3_dendrite.dataclasses.models import SwapTransactionList

load_dotenv()


class DbsyncBackend(AbstractBackend):
    """Concrete implementation of AbstractBackend for db-sync.

    This class provides methods to interact with a Cardano db-sync database
    for retrieving blockchain data.
    """

    def __init__(self) -> None:
        """Initialize the DbsyncBackend with database connection details."""
        self.lock = Lock()
        self.POOL = None
        self.DBSYNC_USER = os.environ.get("DBSYNC_USER", None)
        self.DBSYNC_PASS = os.environ.get("DBSYNC_PASS", None)
        self.DBSYNC_HOST = os.environ.get("DBSYNC_HOST", None)
        self.DBSYNC_PORT = os.environ.get("DBSYNC_PORT", None)
        self.DBSYNC_DB_NAME = os.environ.get("DBSYNC_DB_NAME", None)

    def get_dbsync_pool(self) -> psycopg_pool.ConnectionPool:
        """Get or create a connection pool for the db-sync database.

        Returns:
            psycopg_pool.ConnectionPool: A connection pool for database operations.
        """
        with self.lock:
            if self.POOL is None:
                conninfo = (
                    f"host={self.DBSYNC_HOST} "
                    + f"port={self.DBSYNC_PORT} "
                    + f"dbname={self.DBSYNC_DB_NAME} "
                    + f"user={self.DBSYNC_USER} "
                    + f"password={self.DBSYNC_PASS}"
                )
                self.POOL = psycopg_pool.ConnectionPool(
                    conninfo=conninfo,
                    open=False,
                    min_size=1,
                    max_size=10,
                    max_idle=10,
                    reconnect_timeout=30,  # Increased from 10 to 30
                    max_lifetime=60,
                    check=psycopg_pool.ConnectionPool.check_connection,
                )
                try:
                    if self.POOL is None:
                        raise ValueError("Connection pool has not been initialized.")

                    self.POOL.open()
                    self.POOL.wait(timeout=60.0)  # Increased from 30 to 60 seconds
                except PoolTimeout as e:
                    logging.error(
                        f"Database connection pool initialization timed out: {e}",
                    )
                    logging.error(
                        f"Connection info: host={self.DBSYNC_HOST}, "
                        + f"port={self.DBSYNC_PORT}, "
                        + f"user={self.DBSYNC_USER}",
                    )
                    raise
                except Exception as e:
                    logging.error(f"Error initializing database connection pool: {e}")
                    raise
        return self.POOL

    def db_query(self, query: str, args: dict | None = None) -> list[dict]:
        """Execute a database query using the connection pool.

        Args:
            query (str): The SQL query to execute.
            args (Optional[tuple]): Arguments to be used with the query.

        Returns:
            List[tuple]: The query results.
        """
        with self.get_dbsync_pool().connection() as conn, conn.cursor(
            row_factory=dict_row,
        ) as cursor:
            cursor.execute(query, args)
            return cursor.fetchall()

    def get_pool_utxos(
        self,
        addresses: list[str],
        assets: list[str] | None = None,
        limit: int = 1000,
        page: int = 0,
        historical: bool = True,
    ) -> PoolStateList:
        """Get transactions by policy or address.

        Args:
            addresses: A list of addresses for pool or order contracts
            assets: A list of assets used to filter utxos. Defaults to None.
            limit: Number of values to return. Defaults to 1000.
            page: Page of pools to return. Defaults to 0.
            historical: If False, returns current pool states. Defaults to True.

        Returns:
            A list of pool states.
        """
        # Use the pool selector to format the output
        datum_selector = PoolSelector.select()

        # Get txo from pool script address
        datum_selector += """FROM (
    SELECT *
    FROM tx_out
    WHERE tx_out.payment_cred = ANY(%(addresses)b)
) as txo"""

        # If assets are specified, select assets
        if assets is not None:
            datum_selector += """
LEFT JOIN ma_tx_out mtxo ON mtxo.tx_out_id = txo.id
LEFT JOIN multi_asset ma ON ma.id = mtxo.ident"""

        datum_selector += """
LEFT JOIN tx ON txo.tx_id = tx.id
LEFT JOIN datum ON txo.data_hash = datum.hash
LEFT JOIN block ON tx.block_id = block.id
WHERE datum.hash IS NOT NULL"""

        if not historical:
            datum_selector += """
AND txo.consumed_by_tx_id IS NULL"""

        if assets is not None:
            datum_selector += """
AND ma.policy = ANY(%(policies)b) AND ma.name = ANY(%(names)b)"""

        datum_selector += """
LIMIT %(limit)s
OFFSET %(offset)s"""

        values = {
            "limit": limit,
            "offset": page * limit,
            "addresses": [Address.decode(a).payment_part.payload for a in addresses],
        }
        if assets is not None:
            values.update({"policies": [bytes.fromhex(p[:56]) for p in assets]})
            values.update({"names": [bytes.fromhex(p[56:]) for p in assets]})

        r = self.db_query(datum_selector, values)

        return PoolSelector.parse(r)

    def get_pool_in_tx(
        self,
        tx_hash: str,
        addresses: list[str],
        assets: list[str] | None = None,
    ) -> PoolStateList:
        """Get transactions by policy or address."""
        # Use the pool selector to format the output
        datum_selector = PoolSelector.select()

        datum_selector += """FROM (
    SELECT *
    FROM tx_out
    WHERE tx_out.payment_cred = ANY(%(addresses)b)
) as txo"""

        # If assets are specified, select assets
        if assets is not None:
            datum_selector += """
LEFT JOIN ma_tx_out mtxo ON mtxo.tx_out_id = txo.id
LEFT JOIN multi_asset ma ON ma.id = mtxo.ident"""

        datum_selector += """
LEFT JOIN tx ON txo.tx_id = tx.id
LEFT JOIN datum ON txo.data_hash = datum.hash
LEFT JOIN block ON tx.block_id = block.id
WHERE datum.hash IS NOT NULL AND tx.hash = DECODE(%(tx_hash)s, 'hex')"""

        if assets is not None:
            datum_selector += """
AND ma.policy = ANY(%(policies)b) AND ma.name = ANY(%(names)b)"""

        values = {
            "tx_hash": tx_hash,
            "addresses": [Address.decode(a).payment_part.payload for a in addresses],
        }
        if assets is not None:
            values.update({"policies": [bytes.fromhex(p[:56]) for p in assets]})
            values.update({"names": [bytes.fromhex(p[56:]) for p in assets]})

        r = self.db_query(datum_selector, values)

        return PoolSelector.parse(r)

    def last_block(self, last_n_blocks: int = 2) -> BlockList:
        """Get the last n blocks."""
        r = self.db_query(
            """
    SELECT epoch_slot_no,
    block_no,
    tx_count,
    EXTRACT(
        epoch
        FROM block.time
    )::INTEGER AS "block_time"
    FROM block
    WHERE block_no IS NOT null
    ORDER BY block_no DESC
    LIMIT %(last_n_blocks)s""",
            {"last_n_blocks": last_n_blocks},
        )
        return BlockList.model_validate(r)

    def get_pool_utxos_in_block(self, block_no: int) -> PoolStateList:
        """Get pool utxos in block."""
        # Use this for gathering all assets for multiple addresses
        datum_selector = (
            PoolSelector.select()
            + """
    FROM tx_out txo
    LEFT JOIN tx ON txo.tx_id = tx.id
    LEFT JOIN datum ON txo.data_hash = datum.hash
    LEFT JOIN block ON tx.block_id = block.id
    WHERE block.block_no = %(block_no)s AND datum.hash IS NOT NULL
    """
        )
        r = self.db_query(datum_selector, {"block_no": block_no})

        return PoolSelector.parse(r)

    def get_script_from_address(self, address: Address) -> ScriptReference:
        """Get a reference script from an address."""
        query = UTxOSelector.select()

        query += """
FROM script s
LEFT JOIN tx_out ON s.id = tx_out.reference_script_id
LEFT JOIN tx ON tx.id = tx_out.tx_id
LEFT JOIN datum ON tx_out.inline_datum_id = datum.id
LEFT JOIN block on block.id = tx.block_id
WHERE s.hash = %(address)b AND tx_out.consumed_by_tx_id IS NULL
ORDER BY block.time DESC
LIMIT 1
"""
        r = self.db_query(query, {"address": address.payment_part.payload})

        if r[0]["assets"] is not None and r[0]["assets"][0]["lovelace"] is None:
            r[0]["assets"] = None

        return UTxOSelector.parse(r[0])

    def get_datum_from_address(
        self,
        address: Address,
        asset: str | None = None,
    ) -> ScriptReference | None:
        """Get a reference datum from an address."""
        kwargs = {"address": address.payment_part.payload}

        if asset is not None:
            kwargs.update(
                {
                    "policy": bytes.fromhex(asset[:56]),
                    "name": bytes.fromhex(asset[56:]),
                },
            )

        query = UTxOSelector.select()

        query += """
FROM tx_out
LEFT JOIN ma_tx_out mtxo ON mtxo.tx_out_id = tx_out.id
LEFT JOIN multi_asset ma ON ma.id = mtxo.ident
LEFT JOIN tx ON tx.id = tx_out.tx_id
LEFT JOIN datum ON tx_out.inline_datum_id = datum.id
LEFT JOIN block on block.id = tx.block_id
LEFT JOIN script s ON s.id = tx_out.reference_script_id
WHERE tx_out.payment_cred = %(address)b AND tx_out.consumed_by_tx_id IS NULL"""

        if asset is not None:
            query += """
AND policy = %(policy)b AND name = %(name)b
"""

        query += """
AND tx_out.inline_datum_id IS NOT NULL
ORDER BY block.time DESC
LIMIT 1
"""
        r = self.db_query(query, kwargs)

        if r[0]["assets"] is not None and r[0]["assets"][0]["lovelace"] is None:
            r[0]["assets"] = None

        return UTxOSelector.parse(r[0])

    def get_historical_order_utxos(
        self,
        stake_addresses: list[str],
        after_time: datetime | int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get historical orders at an order submission address."""
        if isinstance(after_time, int):
            after_time = datetime.fromtimestamp(after_time)

        utxo_selector = OrderSelector.select()

        utxo_selector += """FROM (
    SELECT *
    FROM tx_out txo
    WHERE txo.payment_cred = ANY(%(addresses)b) AND txo.data_hash IS NOT NULL
) txo_stake
LEFT JOIN tx ON tx.id = txo_stake.tx_id
LEFT JOIN block ON tx.block_id = block.id
LEFT JOIN datum ON txo_stake.data_hash = datum.hash
LEFT JOIN (
    SELECT tx.hash AS "tx_hash",
    txo.index AS "tx_index",
    txo.value,
    txo.id as "tx_id",
    block.hash AS "block_hash",
    block.time AS "block_time",
    block.block_no,
    tx.block_index AS "block_index",
    tx_in.tx_id as "tx_out_id",
    tx_in.index as "tx_out_index",
    txo.inline_datum_id,
    txo.reference_script_id,
    txo.address,
    datum.hash as "datum_hash",
    datum.bytes as "datum_bytes"
    FROM tx_out tx_in
    LEFT JOIN tx ON tx.id = tx_in.consumed_by_tx_id
    LEFT JOIN tx_out txo ON tx.id = txo.tx_id
    LEFT JOIN block ON tx.block_id = block.id
    LEFT JOIN datum ON txo.data_hash = datum.hash
) txo_output ON txo_output.tx_out_id = txo_stake.tx_id
    AND txo_output.tx_out_index = txo_stake.index
WHERE datum.hash IS NOT NULL"""

        if after_time is not None:
            utxo_selector += """
AND block.time >= %(after_time)s"""

        utxo_selector += """
ORDER BY tx.id ASC
LIMIT %(limit)s
OFFSET %(offset)s"""

        r = self.db_query(
            utxo_selector,
            {
                "addresses": [
                    Address.decode(a).payment_part.payload for a in stake_addresses
                ],
                "limit": limit,
                "offset": page * limit,
                "after_time": (
                    None
                    if after_time is None
                    else after_time.strftime("%Y-%m-%d %H:%M:%S")
                ),
            },
        )

        return OrderSelector.parse(r)

    def get_order_utxos_by_block_or_tx(
        self,
        stake_addresses: list[str],
        out_tx_hash: list[str] | None = None,
        in_tx_hash: list[str] | None = None,
        block_no: int | None = None,
        after_block: int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get order UTxOs by either block number or tx hash."""
        utxo_selector = """
    SELECT (
        SELECT array_agg(DISTINCT txo.address)
        FROM tx_out txo
        WHERE txo.consumed_by_tx_id = txo_stake.tx_id
    ) AS "submit_address_inputs",
    txo_stake.address as "submit_address_stake",
    ENCODE(txo_stake.tx_hash, 'hex') as "submit_tx_hash",
    txo_stake.index as "submit_tx_index",
    ENCODE(txo_stake.block_hash,'hex') as "submit_block_hash",
    EXTRACT(
        epoch
        FROM txo_stake.block_time
    )::INTEGER AS "submit_block_time",
    txo_stake.block_index AS "submit_block_index",
    (
        SELECT array_agg(tx_metadata.json)
        FROM tx_metadata
        WHERE txo_stake.tx_id = tx_metadata.tx_id
    ) AS "submit_metadata",
    COALESCE(
        json_build_object('lovelace',txo_stake.value::TEXT)::jsonb || (
            SELECT json_agg(
                json_build_object(
                    CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                    mto.quantity::TEXT
                )
            )
            FROM ma_tx_out mto
            JOIN multi_asset ma ON (mto.ident = ma.id)
            WHERE mto.tx_out_id = txo_stake.id
        )::jsonb,
        jsonb_build_array(json_build_object('lovelace',txo_stake.value::TEXT)::jsonb)
    ) AS "submit_assets",
    ENCODE(txo_stake.datum_hash,'hex') as "submit_datum_hash",
    ENCODE(txo_stake.datum_bytes,'hex') as "submit_datum_cbor",
    txo_output.address,
    ENCODE(txo_output.tx_hash, 'hex') as "tx_hash",
    txo_output.tx_index as "tx_index",
    EXTRACT(
        epoch
        FROM txo_output.block_time
    )::INTEGER AS "block_time",
    txo_output.block_index AS "block_index",
    ENCODE(txo_output.block_hash,'hex') AS "block_hash",
    ENCODE(txo_output.datum_hash, 'hex') AS "datum_hash",
    ENCODE(txo_output.datum_bytes, 'hex') AS "datum_cbor",
    COALESCE(
        json_build_object('lovelace',txo_output.value::TEXT)::jsonb || (
            SELECT json_agg(
                json_build_object(
                    CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                    mto.quantity::TEXT
                )
            )
            FROM ma_tx_out mto
            JOIN multi_asset ma ON (mto.ident = ma.id)
            WHERE mto.tx_out_id = txo_output.tx_id
        )::jsonb,
        jsonb_build_array(json_build_object('lovelace',txo_output.value::TEXT)::jsonb)
    ) AS "assets",
    (
        txo_output.inline_datum_id IS NOT NULL OR
        txo_output.reference_script_id IS NOT NULL
    ) as "plutus_v2"
    """

        utxo_selector += """FROM (
        SELECT DISTINCT txo.tx_id,
        txo.id,
        txo.index,
        txo.value,
        txo.data_hash,
        txo.address,
        tx.hash as "tx_hash",
        tx.block_index,
        block.hash as "block_hash",
        block.time as "block_time",
        datum.hash as "datum_hash",
        datum.bytes as "datum_bytes"
        FROM tx_out txo
        LEFT JOIN tx ON tx.id = txo.tx_id
        LEFT JOIN block ON tx.block_id = block.id
        LEFT JOIN datum ON txo.data_hash = datum.hash
        LEFT JOIN tx tx_in_ref ON txo.consumed_by_tx_id = tx_in_ref.id
        WHERE txo.payment_cred = ANY(%(addresses)b) AND txo.data_hash IS NOT NULL"""

        if out_tx_hash is not None:
            utxo_selector += """
        AND tx_in_ref.hash = ANY(%(out_tx_hash)b)"""
        elif in_tx_hash is not None:
            utxo_selector += """
        AND tx.hash = ANY(%(in_tx_hash)b)"""

        if block_no is not None:
            utxo_selector += """
        AND block.block_no = %(block_no)s"""
        elif after_block is not None:
            utxo_selector += """
        AND block.block_no >= %(after_block)s"""

        utxo_selector += """
    ) txo_stake
    LEFT JOIN (
        SELECT tx.hash AS "tx_hash",
        txo.index AS "tx_index",
        txo.value,
        txo.id as "tx_id",
        block.hash AS "block_hash",
        block.time AS "block_time",
        block.block_no,
        tx.block_index AS "block_index",
        tx_in.tx_id as "tx_out_id",
        tx_in.index as "tx_out_index",
        txo.inline_datum_id,
        txo.reference_script_id,
        txo.address,
        datum.hash as "datum_hash",
        datum.bytes as "datum_bytes"
        FROM tx_out tx_in
        LEFT JOIN tx ON tx.id = tx_in.consumed_by_tx_id
        LEFT JOIN tx_out txo ON tx.id = txo.tx_id
        LEFT JOIN block ON tx.block_id = block.id
        LEFT JOIN datum ON txo.data_hash = datum.hash
    ) txo_output ON txo_output.tx_out_id = txo_stake.tx_id
        AND txo_output.tx_out_index = txo_stake.index
    WHERE txo_stake.datum_hash IS NOT NULL
    ORDER BY txo_stake.tx_id ASC
    LIMIT %(limit)s
    OFFSET %(offset)s"""

        r = self.db_query(
            utxo_selector,
            {
                "addresses": [
                    Address.decode(a).payment_part.payload for a in stake_addresses
                ],
                "limit": limit,
                "offset": page * limit,
                "block_no": block_no,
                "after_block": after_block,
                "out_tx_hash": (
                    None
                    if out_tx_hash is None
                    else [bytes.fromhex(h) for h in out_tx_hash]
                ),
                "in_tx_hash": (
                    None
                    if in_tx_hash is None
                    else [bytes.fromhex(h) for h in in_tx_hash]
                ),
            },
        )

        return OrderSelector.parse(r)

    def get_cancel_utxos(
        self,
        stake_addresses: list[str],
        block_no: int | None = None,
        after_time: datetime | int | None = None,
        limit: int = 1000,
        page: int = 0,
    ) -> SwapTransactionList:
        """Get order cancel UTxOs."""
        if isinstance(after_time, int):
            after_time = datetime.fromtimestamp(after_time)

        utxo_selector = """
SELECT (
    SELECT array_agg(DISTINCT tx_out.address)
    FROM tx_out
    WHERE tx_out.consumed_by_tx_id = txo.tx_id
) AS "submit_address_inputs",
txo.address as "submit_address_stake",
ENCODE(tx.hash, 'hex') as "submit_tx_hash",
txo.index as "submit_tx_index",
ENCODE(block.hash,'hex') as "submit_block_hash",
EXTRACT(
    epoch
    FROM block.time
)::INTEGER AS "submit_block_time",
tx.block_index AS "submit_block_index",
(
    SELECT array_agg(tx_metadata.json)
    FROM tx_metadata
    WHERE txo.tx_id = tx_metadata.tx_id
) AS "submit_metadata",
COALESCE(
    json_build_object('lovelace',txo.value::TEXT)::jsonb || (
        SELECT json_agg(
            json_build_object(
                CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                mto.quantity::TEXT
            )
        )
        FROM ma_tx_out mto
        JOIN multi_asset ma ON (mto.ident = ma.id)
        WHERE mto.tx_out_id = txo.id
    )::jsonb,
    jsonb_build_array(json_build_object('lovelace',txo.value::TEXT)::jsonb)
) AS "submit_assets",
ENCODE(datum.hash,'hex') as "submit_datum_hash",
ENCODE(datum.bytes,'hex') as "submit_datum_cbor",
txo_output.address,
ENCODE(txo_output.tx_hash, 'hex') as "tx_hash",
txo_output.tx_index as "tx_index",
EXTRACT(
    epoch
    FROM txo_output.block_time
)::INTEGER AS "block_time",
txo_output.block_index AS "block_index",
ENCODE(txo_output.block_hash,'hex') AS "block_hash",
ENCODE(txo_output.datum_hash, 'hex') AS "datum_hash",
ENCODE(txo_output.datum_bytes, 'hex') AS "datum_cbor",
COALESCE(
    json_build_object('lovelace',txo_output.value::TEXT)::jsonb || (
        SELECT json_agg(
            json_build_object(
                CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                mto.quantity::TEXT
            )
        )
        FROM ma_tx_out mto
        JOIN multi_asset ma ON (mto.ident = ma.id)
        WHERE mto.tx_out_id = txo_output.tx_id
    )::jsonb,
    jsonb_build_array(json_build_object('lovelace',txo_output.value::TEXT)::jsonb)
) AS "assets",
(txo_output.inline_datum_id IS NOT NULL OR txo_output.reference_script_id IS NOT NULL)
    AS "plutus_v2"
"""

        utxo_selector += """FROM (
    SELECT tx.hash AS "tx_hash",
    txo.index AS "tx_index",
    txo.value,
    txo.id as "tx_id",
    block.hash AS "block_hash",
    block.time AS "block_time",
    block.block_no,
    tx.block_index AS "block_index",
    tx_in.tx_id as "tx_out_id",
    tx_in.index as "tx_out_index",
    txo.inline_datum_id,
    txo.reference_script_id,
    txo.address,
    datum.hash as "datum_hash",
    datum.bytes as "datum_bytes"
    FROM tx_out tx_in
    LEFT JOIN tx ON tx.id = tx_in.tx_id
    LEFT JOIN tx_out txo ON tx.id = txo.tx_id
    LEFT JOIN block ON tx.block_id = block.id
    LEFT JOIN datum ON txo.data_hash = datum.hash"""

        if after_time is not None:
            utxo_selector += """
    WHERE block.time >= %(after_time)s"""
        elif block_no is not None:
            utxo_selector += """
    WHERE block.block_no = %(block_no)s"""
        else:
            raise ValueError("Either after_time or block_no should be defined.")

        utxo_selector += """
    GROUP BY tx.hash, txo.value, txo.id, block.hash, block.time, block.block_no,
    tx.block_index, tx_in.tx_id, tx_in.index, txo.inline_datum_id,
    txo.reference_script_id, txo.address, datum.hash, datum.bytes
    HAVING COUNT(DISTINCT txo.address) = 1
) txo_output
LEFT JOIN tx_out txo ON txo.tx_id = txo_output.tx_out_id
    AND txo_output.tx_out_index = txo.index
    AND txo.payment_cred = ANY(%(addresses)b)
    AND txo.data_hash IS NOT NULL
LEFT JOIN tx ON tx.id = txo.tx_id
LEFT JOIN block ON tx.block_id = block.id
LEFT JOIN datum ON txo.data_hash = datum.hash
LEFT JOIN tx tx_in_ref ON txo.tx_id = tx_in_ref.id
WHERE txo.id IS NOT NULL
ORDER BY txo.tx_id ASC
LIMIT %(limit)s
OFFSET %(offset)s"""

        r = self.db_query(
            utxo_selector,
            {
                "addresses": [
                    Address.decode(a).payment_part.payload for a in stake_addresses
                ],
                "limit": limit,
                "offset": page * limit,
                "after_time": (
                    None
                    if after_time is None
                    else after_time.strftime("%Y-%m-%d %H:%M:%S")
                ),
                "block_no": block_no,
            },
        )

        return SwapTransactionList.model_validate(r)

    def get_axo_target(
        self,
        assets: Assets,
        block_time: datetime | None = None,
    ) -> str | None:
        """Get the target address for the given asset."""
        query = """
    SELECT DISTINCT txo.address, block.time
    FROM (
        SELECT tx.id, tx.block_id
        FROM tx_out txo
        LEFT JOIN tx ON tx.id = txo.tx_id
        WHERE txo.payment_cred = DECODE(
            '55ff0e63efa0694e8065122c552e80c7b51768b7f20917af25752a7c', 'hex'
        )
    ) as tx
    LEFT JOIN block ON block.id = tx.block_id
    LEFT JOIN tx_out txo ON tx.id = txo.tx_id
    LEFT JOIN ma_tx_out mtxo on txo.id = mtxo.tx_out_id
    LEFT JOIN multi_asset ma ON ma.id = mtxo.ident
    WHERE ma.policy = %(policy)b AND ma.name = %(name)b
    AND txo.payment_cred != DECODE(
        '55ff0e63efa0694e8065122c552e80c7b51768b7f20917af25752a7c', 'hex'
    )"""

        if block_time is not None:
            query += """
    AND block.time <= %(block_time)s"""

        query += """
    ORDER BY block.time DESC"""

        policy = bytes.fromhex(assets.unit()[:56])
        name = bytes.fromhex(assets.unit()[56:])
        r = self.db_query(
            query,
            {
                "policy": policy,
                "name": name,
                "block_time": (
                    None
                    if block_time is None
                    else block_time.strftime("%Y-%m-%d %H:%M:%S")
                ),
            },
        )

        if len(r) == 0:
            return None

        return r[0]["address"]
